from keras.utils import to_categorical


class Performance():

    '''A modified fscore where the predictions can be
    single-label, multi-label, or multi-category, and
    each will be treated in a way that allows comparison
    (apples-to-apples) between different prediction tasks.'''

    def __init__(self, y_pred, y_val, shape, y_max):

        self.y_pred = y_pred
        self.y_val = y_val
        self.shape = shape

        self.classes = y_max + 1

        if self.shape == 'binary_class':
            self.binary_class()
        elif self.shape == 'multi_class':
            self.multi_class()
        elif self.shape == 'multi_label':
            self.multi_label()

        self.trues_and_falses()
        self.f1score()
        self.balance()
        self.one_rule()
        self.zero_rule()

    def multi_class(self):

        '''For one-hot encoded'''

        self.y_pred = self.y_pred.flatten('F')
        self.y_val = self.y_val.flatten('F')

    def binary_class(self):

        '''For single column, single label'''

        return

    def multi_label(self):

        '''For many labels in a single column'''

        self.y_pred = to_categorical(self.y_pred, num_classes=self.classes)
        self.y_val = to_categorical(self.y_val, num_classes=self.classes)

        self.multi_class()

    def f1score(self):

        '''Computes fscore when possible'''

        if sum(self.y_pred) == len(self.y_pred):
            if sum(self.y_val) != len(self.y_val):
                self.result = '_warning_all_ones_'
                return
        elif sum(self.y_pred) == 0:
            if sum(self.y_val) != 0:
                self.result = '_warning_all_zeros_'
            elif sum(self.y_val) == 0:
                self.result = 1
            return
        try:
            self.precision = self.tp / (self.tp + self.fp)
            1 / self.precision
        except ZeroDivisionError:
            self.result = '_warning_no_true_positive'
            return

        try:
            self.recall = self.tp / (self.tp + self.fn)
            1 / self.recall
        except ZeroDivisionError:
            self.result = '_warning_3'
            return

        try:
            f1 = 2 * ((self.precision * self.recall) /
                      (self.precision + self.recall))
            self.result = f1
        except ZeroDivisionError:
            return

    def trues_and_falses(self):

        '''Returns tp, tn, fp, and fn values'''

        self.pos = sum(self.y_val)
        self.neg = len(self.y_val) - sum(self.y_val)

        self.tp, self.tn, self.fp, self.fn = 0, 0, 0, 0

        # then we iterate through the predictions
        for i in range(len(self.y_pred)):
            if self.y_pred[i] == 1 and self.y_val[i] == 1:
                self.tp += 1
            elif self.y_pred[i] == 1 and self.y_val[i] == 0:
                self.fp += 1
            elif self.y_pred[i] == 0 and self.y_val[i] == 0:
                self.tn += 1
            elif self.y_pred[i] == 0 and self.y_val[i] == 1:
                self.fn += 1

    def balance(self):

        '''Counts the balance between 0s and 1s'''

        self.balance = self.pos / (self.pos + self.neg)

    def zero_rule(self):

        '''Returns accuracy for all 0s'''

        self.zero_rule = 1 - self.balance

    def one_rule(self):

        '''Returns accuracy for all 1s'''

        self.one_rule = self.balance



from .scan.Scan import Scan
from .commands.reporting import Reporting
from .commands.predict import Predict
from .commands.deploy import Deploy
from .commands.evaluate import Evaluate
from. commands.restore import Restore

from .metrics.performance import Performance
from .examples import datasets, params
import astetik as plots
from kerasplotlib import TrainingLog as live

# del parameters, utils, scan
# del Performance, reporting, reducers, metrics, examples

__version__ = "0.4.3"



from keras import backend as K

from ..parameters.round_params import round_params
from ..utils.results import create_header
from ..metrics.entropy import epoch_entropy
from ..model.ingest_model import ingest_model
from ..metrics.score_model import get_score
from ..utils.logging import write_log
from ..utils.results import run_round_results, save_result
from ..reducers.reduce_run import reduce_run
from ..utils.exceptions import TalosReturnError, TalosTypeError


def scan_round(self):

    '''The main operational function that manages the experiment
    on the level of execution of each round.'''

    # determine the parameters for the particular execution
    self.round_params = round_params(self)

    # print round params
    if self.print_params is True:
        print(self.round_params)

    # compile the model
    try:
        _hr_out, self.keras_model = ingest_model(self)
    except TypeError as err:
        if err.args[0] == "unsupported operand type(s) for +: 'int' and 'numpy.str_'":
            raise TalosTypeError("Activation should be as object and not string in params")
        else:
            raise TalosReturnError("Make sure that input model returns 'out, model' where out is history object from model.fit()")

    # create log and other stats
    try:
        self.epoch_entropy.append(epoch_entropy((_hr_out)))
    except (TypeError, AttributeError):
        raise TalosReturnError("Make sure that input model returns in the order 'out, model'")

    if self.round_counter == 0:
        _for_header = create_header(self, _hr_out)
        self.result.append(_for_header)
        save_result(self)

    _hr_out = run_round_results(self, _hr_out)
    self._val_score = get_score(self)
    write_log(self)
    self.result.append(_hr_out)
    save_result(self)

    # apply reduction
    if self.reduction_method is not None:
        if (self.round_counter + 1) % self.reduction_interval == 0:
            len_before_reduce = len(self.param_log)
            self = reduce_run(self)
            total_reduced = len_before_reduce - len(self.param_log)
            # update the progress bar
            self.pbar.update(total_reduced)

    # save model and weights
    self.saved_models.append(self.keras_model.to_json())
    self.saved_weights.append(self.keras_model.get_weights())

    # clear tensorflow sessions (maybe)
    if self.clear_tf_session is True:
        K.clear_session()
    self.round_counter += 1

    return self



from .performance import Performance
from numpy import nan


def _functional_predict(model, x, batch_size=32, verbose=0):

    '''Supports functional model class from Keras'''

    proba = model.predict(x,
                          batch_size=batch_size,
                          verbose=verbose)

    if proba.shape[-1] > 1:
        return proba.argmax(axis=-1)
    else:
        return (proba > 0.5).astype('int32')


def get_score(self):

    '''SCORE MODEL CONFIGURATION

    Initiates the model scoring process. Currently works
    only for binary and categorical predictions. The model scoring
    currently is only used in the Talos master log, and is not
    available in the experiment log or reports.

    TODO: the exception handling needs to become metric based.

    '''

    try:
        # for functional model experimental support
        if self.functional_model:
            y_pred = _functional_predict(self.keras_model, self.x_val)
        # all other cases
        else:
            y_pred = self.keras_model.predict_classes(self.x_val)

        return Performance(y_pred, self.y_val, self.shape, self.y_max).result

    except TypeError:
        return nan



from scipy.stats import entropy
from numpy import nan


def epoch_entropy(history):

    '''MEASURE EPOCH ENTROPY

    BINARY/CATEGORICAL:

    Measures the KL divergence of the acc and loss results
    per epoch of a given permutation.

    CONTINUOUS:

    Measures shannon entropy for loss.

    # TODO Right now this does not handle all cases well and needs
      to be thought about properly.
    '''

    keys = list(history.history.keys())
    no_of_items = len(keys)

    if no_of_items == 1:
        if 'loss' in keys:
            loss_entropy = entropy(history.history['loss'])
            acc_entropy = nan
        else:
            loss_entropy = nan
            acc_entropy = nan

    elif no_of_items == 2:
        if 'acc' in keys and 'loss' in keys:
            loss_entropy = entropy(history.history['loss'])
            acc_entropy = entropy(history.history['acc'])
        else:
            loss_entropy = nan
            acc_entropy = nan

    elif no_of_items >= 4:
        if 'acc' in keys:
            acc_entropy = entropy(history.history['val_acc'],
                                  history.history['acc'])
        else:
            acc_entropy = nan

        if 'loss' in keys:
            loss_entropy = entropy(history.history['val_loss'],
                                   history.history['loss'])
        else:
            loss_entropy = nan

    return [acc_entropy, loss_entropy]



import math as m


def shapes(params, last_neuron):

    '''Handles shapes for the case where hidden layers are used
    as an optimization parameter. Available shapes are:

    'funnel',
    'long_funnel',
    'rhombus',
    'brick',
    'diamond',
    'hexagon',
    'triangle',
    'stairs'

    '''
    layers = params['hidden_layers']
    shape = params['shapes']
    neuron_max = params['first_neuron']
    neuron_previous = neuron_max
    neuron_last = last_neuron

    neuron_count = []
    l = []

    if shape == 'funnel':

        neuron_count.append(neuron_max)

        for i in range(layers - 1):
            neuron_count.append((neuron_previous + neuron_last) / 2)
            neuron_previous = neuron_count[i+1]

    if shape == 'rhombus':

        neuron_count.append(1)

        a = neuron_max * 2 - 1
        k = 0

        if a < layers:
            k = layers - a
            val = layers - k

        else:
            val = layers

        if (val % 2) == 0:

            n = int((val - 2) / 2)

            for i in range(int(n)):
                neuron_count.append(int(neuron_max * (i+1) * 2 / val))

            l = neuron_count[::-1]
            neuron_count.append(neuron_max)

            for i in range(int(n)):
                neuron_count.append(l[i])

        else:

            n = (val - 1) / 2

            for i in range(int(n) - 1):

                neuron_count.append(int(neuron_max * (3 + 2 * i) / val))

            l = neuron_count[::-1]
            neuron_count.append(neuron_max)

            for i in range(int(n)):
                neuron_count.append(l[i])

        if k != 0:
            for i in range(int(n)):
                neuron_count.append(1)

    if shape == 'long_funnel':

        n = layers / 2

        if (layers % 2 == 0):

            for i in range(int(n)):
                neuron_count.append(neuron_max)

        elif (layers % 2 == 1):

            for i in range(int(n) + 1):
                neuron_count.append(neuron_max)

        for i in range(int(n)):
            neuron_count.append((neuron_previous + neuron_last) / 2)
            neuron_previous = neuron_count[i + int(n)]

    if shape == 'brick':

        for i in range(layers):
            neuron_count.append(neuron_max)

    if shape == 'diamond':

        n = layers / 2

        if (layers % 2 == 0):

            for i in range(int(n) - 1):
                neuron_count.append(int(m.ceil(neuron_max * layers + i) / 2.0) / (layers - 1))
            neuron_count.append(neuron_max)

        else:

            for i in range(int(n)):
                neuron_count.append(int(m.ceil(neuron_max * layers + i) / 2.0) / (layers - 1))
            for i in range(1):
                neuron_count.append(neuron_max)

        for i in range(int(n)):
            neuron_count.append((neuron_previous + neuron_last) / 2)
            neuron_previous = neuron_count[i + int(n)]

    if shape == 'hexagon':

        neuron_count.append(1)

        if (layers == 4):
            for i in range(layers - 1):
                neuron_count.append(neuron_max)

        else:

            n = layers / 3

            for i in range(int(n)):
                l.append((neuron_previous + neuron_last) / 2)
                neuron_previous = l[i]

            l = l[::-1]

            if (layers % 3 == 0):

                for i in range(int(n) - 1):
                    neuron_count.append(l[i])

            if (layers % 3 == 1):

                for i in range(int(n)):
                    neuron_count.append(l[i])

            if (layers % 3 == 2):

                for i in range(int(n)):
                    neuron_count.append(l[i])

                neuron_count.append(neuron_max)

            l = l[::-1]

            for i in range(int(n)):
                neuron_count.append(neuron_max)

            for i in range(int(n)):
                neuron_count.append(l[i])

    if shape == 'triangle':

        neuron_count.append(1)

        for i in range(layers - 2):
            l.append((neuron_previous + neuron_last) / 2)
            neuron_previous = l[i]

        l = l[::-1]

        for i in range(layers - 2):
            neuron_count.append(l[i])

        neuron_count.append(neuron_max)

    if shape == 'stairs':

        if (layers >= 5):
            if (layers % 2 == 1):
                neuron_count.append(neuron_max)

            else:
                for i in range(2):
                    neuron_count.append(neuron_max)

            n = layers / 2

            for i in range(int(n)):
                if neuron_previous >= 3:
                    neuron_previous -= 2
                for j in range(2):
                    neuron_count.append(neuron_previous)

        else:
            for i in range(layers):
                neuron_count.append(neuron_previous)
                neuron_previous -= 1

    return [int(i) for i in neuron_count]



import time
from pandas import Series

attrs_to_keep = ['data', 'x', 'y', 'peak_epochs_df',
                 'random_method', 'grid_downsample',
                 'reduction_interval', 'reduce_loss',
                 'reduction_method', 'reduction_metric',
                 'reduction_threshold', 'reduction_window',
                 'params', 'saved_models', 'saved_weights',
                 'experiment_name']

attrs_final = ['data', 'x', 'y', 'peak_epochs_df',
               'params', 'saved_models', 'saved_weights']


def scan_finish(self):

    # combine entropy tables
    self.peak_epochs_df['acc_epoch'] = [i[0] for i in self.epoch_entropy]
    self.peak_epochs_df['loss_epoch'] = [i[1] for i in self.epoch_entropy]

    # clean the results into a dataframe
    self.data = self.result[self.result.columns[0]].str.split(',', expand=True)
    self.data.columns = self.result.columns[0].split(',')

    # remove redundant columns
    keys = list(self.__dict__.keys())
    for key in keys:
        if key not in attrs_to_keep:
            delattr(self, key)

    # summarize single inputs in dictionary
    out = {}

    for key in list(self.__dict__.keys()):
        if key not in attrs_final:
            out[key] = self.__dict__[key]

    out['complete_time'] = time.strftime('%D/%H:%M')
    out['x_shape'] = self.x.shape
    out['y_shape'] = self.y.shape

    # final cleanup
    keys = list(self.__dict__.keys())
    for key in keys:
        if key not in attrs_final:
            delattr(self, key)

    # add details dictionary as series
    self.details = Series(out)

    return self



from keras import backend as K


def matthews_correlation_acc(y_true, y_pred):

    '''Calculates the Matthews correlation coefficient measure for quality
    of binary classification problems.
    '''

    y_pred_pos = K.round(K.clip(y_pred, 0, 1))
    y_pred_neg = 1 - y_pred_pos

    y_pos = K.round(K.clip(y_true, 0, 1))
    y_neg = 1 - y_pos

    tp = K.sum(y_pos * y_pred_pos)
    tn = K.sum(y_neg * y_pred_neg)

    fp = K.sum(y_neg * y_pred_pos)
    fn = K.sum(y_pos * y_pred_neg)

    numerator = (tp * tn - fp * fn)
    denominator = K.sqrt((tp + fp) * (tp + fn) * (tn + fp) * (tn + fn))

    return numerator / (denominator + K.epsilon())


def precision_acc(y_true, y_pred):

    '''Calculates the precision, a metric for multi-label classification of
    how many selected items are relevant.
    '''

    true_positives = K.sum(K.round(K.clip(y_true * y_pred, 0, 1)))
    predicted_positives = K.sum(K.round(K.clip(y_pred, 0, 1)))
    precision = true_positives / (predicted_positives + K.epsilon())
    return precision


def recall_acc(y_true, y_pred):

    '''Calculates the recall, a metric for multi-label classification of
    how many relevant items are selected.
    '''

    true_positives = K.sum(K.round(K.clip(y_true * y_pred, 0, 1)))
    possible_positives = K.sum(K.round(K.clip(y_true, 0, 1)))
    recall = true_positives / (possible_positives + K.epsilon())
    return recall


def fbeta_score_acc(y_true, y_pred, beta=1):

    '''Calculates the F score, the weighted harmonic mean of precision and recall.
    This is useful for multi-label classification, where input samples can be
    classified as sets of labels. By only using accuracy (precision) a model
    would achieve a perfect score by simply assigning every class to every
    input. In order to avoid this, a metric should penalize incorrect class
    assignments as well (recall). The F-beta score (ranged from 0.0 to 1.0)
    computes this, as a weighted mean of the proportion of correct class
    assignments vs. the proportion of incorrect class assignments.
    With beta = 1, this is equivalent to a F-measure. With beta < 1, assigning
    correct classes becomes more important, and with beta > 1 the metric is
    instead weighted towards penalizing incorrect class assignments.
    '''

    if beta < 0:
        raise ValueError('The lowest choosable beta is zero (only precision).')

    # If there are no true positives, fix the F score at 0 like sklearn.
    if K.sum(K.round(K.clip(y_true, 0, 1))) == 0:
        return 0

    p = precision_acc(y_true, y_pred)
    r = recall_acc(y_true, y_pred)
    bb = beta ** 2
    fbeta_score = (1 + bb) * (p * r) / (bb * p + r + K.epsilon())
    return fbeta_score


def fmeasure_acc(y_true, y_pred):
    '''Calculates the f-measure, the harmonic mean of precision and recall.
    '''
    return fbeta_score_acc(y_true, y_pred, beta=1)



#!/usr/bin/env python

from talos.model import lr_normalizer, early_stopper, hidden_layers

from keras.models import Sequential
from keras.layers import Dropout, Dense

from talos.metrics.keras_metrics import matthews_correlation_acc, precision_acc
from talos.metrics.keras_metrics import recall_acc, fmeasure_acc


def iris_model(x_train, y_train, x_val, y_val, params):

    # note how instead of passing the value, we pass a dictionary entry
    model = Sequential()
    model.add(Dense(params['first_neuron'],
                    input_dim=x_train.shape[1],
                    activation='relu'))

    # same here, just passing a dictionary entry
    model.add(Dropout(params['dropout']))

    # with this call we can create any number of hidden layers
    hidden_layers(model, params, y_train.shape[1])

    # again, instead of the activation name, we have a dictionary entry
    model.add(Dense(y_train.shape[1],
                    activation=params['last_activation']))

    # here are using a learning rate boundary
    model.compile(optimizer=params['optimizer']
                  (lr=lr_normalizer(params['lr'],
                                    params['optimizer'])),
                  loss=params['losses'],
                  metrics=['acc'])

    # here we are also using the early_stopper function for a callback
    out = model.fit(x_train, y_train,
                    batch_size=params['batch_size'],
                    epochs=params['epochs'],
                    verbose=0,
                    validation_data=[x_val, y_val],
                    callbacks=early_stopper(params['epochs'], mode=[1, 1]))

    return out, model


def cervix_model(x_train, y_train, x_val, y_val, params):

    model = Sequential()
    model.add(Dense(params['first_neuron'],
                    input_dim=x_train.shape[1],
                    activation='relu'))

    model.add(Dropout(params['dropout']))

    hidden_layers(model, params, 1)

    model.add(Dense(1, activation=params['last_activation']))

    model.compile(optimizer=params['optimizer']
                  (lr=lr_normalizer(params['lr'],
                                    params['optimizer'])),
                  loss=params['losses'],
                  metrics=['acc',
                           fmeasure_acc,
                           recall_acc,
                           precision_acc,
                           matthews_correlation_acc])

    results = model.fit(x_train, y_train,
                        batch_size=params['batch_size'],
                        epochs=params['epochs'],
                        verbose=0,
                        validation_data=[x_val, y_val],
                        callbacks=early_stopper(params['epochs'],
                                                mode='moderate',
                                                monitor='val_fmeasure'))

    return results, model



from keras.callbacks import EarlyStopping


def early_stopper(epochs,
                  monitor='val_loss',
                  mode='moderate',
                  min_delta=None,
                  patience=None):

    '''EARLY STOP CALLBACK

    Helps prevent wasting time when loss is not becoming
    better. Offers two pre-determined settings 'moderate'
    and 'strict' and allows input of list with two values:

    min_delta = the limit for change at which point flag is raised

    patience = the number of epochs before termination from flag

    '''

    if mode == 'moderate':
        _es_out = EarlyStopping(monitor=monitor,
                                min_delta=0,
                                patience=int(epochs / 10),
                                verbose=0, mode='auto')
    elif mode == 'strict':
        _es_out = EarlyStopping(monitor=monitor,
                                min_delta=0,
                                patience=2,
                                verbose=0, mode='auto')
    elif isinstance(mode, list):
        _es_out = EarlyStopping(monitor=monitor,
                                min_delta=mode[0],
                                patience=mode[1],
                                verbose=0, mode='auto')
    return [_es_out]



from numpy import median, unique, mean


def prediction_type(self):

    try:
        y_cols = self.y.shape[1]
    except IndexError:
        y_cols = 1
    y_max = self.y.max()
    y_uniques = len(unique(self.y))

    if y_cols > 1:
        self._y_type = 'category'
        self._y_range = y_cols
        self._y_format = 'onehot'
    else:
        if y_max == 1:
            self._y_type = 'binary'
            self._y_range = y_cols
            self._y_format = 'single'
        elif mean(self.y) == median(self.y):
            self._y_type = 'category'
            self._y_range = y_uniques
            self._y_format = 'single'
        else:
            self._y_type = 'continuous'
            self._y_num = self.y.max() - self.y.min()
            self._y_format = 'single'

    return self



from time import strftime

from ..utils.validation_split import validation_split
from ..utils.detector import prediction_type
from ..parameters.ParamGrid import ParamGrid
from ..utils.pred_class import classify
from ..utils.last_neuron import last_neuron
from ..utils.logging import debug_logging


TRAIN_VAL_RUNTIME_ERROR_MSG = """
If x_val or y_val is inputted, then the other must be inputted as well.
"""


def scan_prepare(self):

    '''Includes all preparation procedures up until starting the first scan
    through scan_run()'''

    # based on the 'debug' Scan() paramater
    self = debug_logging(self)

    # create the name for the experiment
    if self.dataset_name is None:
        self.dataset_name = strftime('%D%H%M%S').replace('/', '')

    if self.experiment_no is None:
        self.experiment_no = ''

    self.experiment_name = self.dataset_name + '_' + self.experiment_no

    # for the case where x_val or y_val is missing when other is present
    self.custom_val_split = False
    if (self.x_val is not None and self.y_val is None) or \
       (self.x_val is None and self.y_val is not None):
        raise RuntimeError(TRAIN_VAL_RUNTIME_ERROR_MSG)

    elif (self.x_val is not None and self.y_val is not None):
        self.custom_val_split = True

    # create the paramater object and move to self
    self.paramgrid_object = ParamGrid(self)
    self.param_log = self.paramgrid_object.param_log
    self.param_grid = self.paramgrid_object.param_grid
    del self.paramgrid_object

    # creates a reference dictionary for column number to label
    self.param_reference = {}
    for i, col in enumerate(self.params.keys()):
        self.param_reference[col] = i

    self.round_counter = 0
    self.peak_epochs = []
    self.epoch_entropy = []
    self.round_models = []

    # create the data asset
    self.y_max = self.y.max()
    self = validation_split(self)
    self.shape = classify(self.y)
    self.last_neuron = last_neuron(self)

    self._data_len = len(self.x)
    self = prediction_type(self)
    self.result = []

    # model saving
    self.saved_models = []
    self.saved_weights = []

    return self



def last_neuron(self):

    labels = list(set(self.y.flatten('F')))

    try:
        last_neuron = self.y.shape[1]
        return last_neuron
    except IndexError:
        if len(labels) == 2 and max(labels) == 1:
            last_neuron = 1
        elif len(labels) == 2 and max(labels) > 1:
            last_neuron = 3
        elif len(labels) > 2:
            last_neuron = len(labels)

    return last_neuron



from zipfile import ZipFile

from pandas import read_csv
from numpy import load

from talos.utils.load_model import load_model


class Restore:

    '''Utility class for restoring the assets from Deploy()
    package.'''

    def __init__(self, path_to_zip):

        # create paths
        self.path_to_zip = path_to_zip
        self.extract_to = path_to_zip.replace('.zip', '')
        self.package_name = self.extract_to.split('/')[-1]
        self.file_prefix = self.extract_to + '/' + self.package_name

        # extract the zip
        # unpack_archive(self.path_to_zip, self.extract_to)
        z = ZipFile(self.path_to_zip, mode='r')
        z.extractall(self.extract_to)

        # add params dictionary
        self.params = load(self.file_prefix + '_params.npy').item()

        # add experiment details
        self.details = read_csv(self.file_prefix + '_details.txt', header=None)

        # add x data sample
        self.x = read_csv(self.file_prefix + '_x.csv', header=None)

        # add y data sample
        self.y = read_csv(self.file_prefix + '_y.csv', header=None)

        # add model
        self.model = load_model(self.file_prefix + '_model')

        # add results
        self.results = read_csv(self.file_prefix + '_results.csv')
        self.results.drop('Unnamed: 0', axis=1, inplace=True)

        # clean up
        del self.extract_to, self.file_prefix
        del self.package_name, self.path_to_zip



import re


def clean_dict(self):

    '''this is operated from _write_tolog'''

    _cd_out = {}

    for key in self.params.keys():

            s = str(self.params[key])
            if s.startswith('<class'):
                _cd_out[key] = s.strip("<|>|\'").split('.')[2]
            elif s.startswith('<function'):
                _cd_out[key] = s.split()[1]
            elif s.startswith('None'):
                _cd_out[key] = None
            elif len(s.split('.')) > 1:
                _cd_out[key] = self.params[key]
            elif isinstance(self.params[key], int):
                _cd_out[key] = int(s)
            else:
                _cd_out[key] = s

    return _cd_out


def dict_tostr(self, d):

    '''PARSES THE LOG ENTRY'''

    s = "','".join(str(d[key]) for key in d.keys())
    s = re.sub("'", "", s, count=9)
    s = re.sub("$", "'", s)

    # add some values
    try:
        s += ',' + str(round(self._val_score, 3))
    except TypeError:
        s += ',' + self._val_score
        s += ',' + str(self._round_epochs)
        s += ',' + self.shape
        s += ',' + str(self._y_range)
        s += ',' + self._y_format
        s += ',' + str(self.val_split)
        s += ',' + self.dataset_name

    return s


def write_log(self):

    '''this is operated from _hyper_run'''

    _wt_out = clean_dict(self)
    _wt_out = dict_tostr(self, _wt_out)
    self.logfile.write(_wt_out + '\n')


def debug_logging(self):

    if self.debug:
        self.logfile = open('talos.debug.log', 'a')
    else:
        self.logfile_name = self.talos_log_name
        self.logfile = open(self.logfile_name, 'a')

    return self



from numpy import array, argpartition, savetxt
from pandas import DataFrame


def create_header(self, out):

    '''Creates the Header column
    On the first round creates the header columns
    for the experiment output log.
    '''

    _rr_out = []

    _rr_out.append('round_epochs')
    [_rr_out.append(i) for i in list(out.history.keys())]
    [_rr_out.append(key) for key in self.params.keys()]

    self.peak_epochs.append(list(out.history.keys()))

    return ",".join(str(i) for i in _rr_out)


def run_round_results(self, out):

    '''THE MAIN FUNCTION FOR CREATING RESULTS FOR EACH ROUNDself.
    Takes in the history object from model.fit() and handles it.

    NOTE: The epoch level data will be dropped here each round.

    '''

    _rr_out = []

    self._round_epochs = len(list(out.history.values())[0])

    # otherwise proceed to create the value row
    _rr_out.append(self._round_epochs)
    p_epochs = []

    # iterates through the keys and records last or peak for metrics
    for key in out.history.keys():
        t_t = array(out.history[key])

        # this handles metrics (NOTE: 'acc' have to be in metric name)
        if 'acc' in key:
            best_epoch = argpartition(t_t, len(t_t) - 1)[-1]

        # this handles losses (takes minimum value epoch)
        else:
            best_epoch = argpartition(t_t, len(t_t) - 1)[0]

        if self.last_epoch_value:
            value_to_report = out.history[key][-1]
        else:
            value_to_report = array(out.history[key])[best_epoch]

        _rr_out.append(value_to_report)
        p_epochs.append(best_epoch)

        # this takes care of the separate entity with just peak epoch data
    self.peak_epochs.append(p_epochs)

    for key in self.round_params.keys():
        _rr_out.append(self.round_params[key])

    return ",".join(str(i) for i in _rr_out)


def save_result(self):
    '''SAVES THE RESULTS/PARAMETERS TO A CSV SPECIFIC TO THE EXPERIMENT'''

    savetxt(self.experiment_name + '.csv',
            self.result,
            fmt='%s',
            delimiter=',')


def result_todf(self):
    '''ADDS A DATAFRAME VERSION OF THE RESULTS TO THE CLASS OBJECT'''

    self.result = DataFrame(self.result)
    self.result.columns = self.result.iloc[0]
    self.result = self.result.drop(0)

    return self


def peak_epochs_todf(self):

    return DataFrame(self.peak_epochs, columns=self.peak_epochs[0]).drop(0)



import pandas as pd

from ..metrics.names import metric_names


def reduce_prepare(self):

    '''
    Preparation procedures for applying a reduction algorithm.
    '''

    # load the data from the experiment log
    self.data = pd.read_csv(self.experiment_name + '.csv')
    self.names = metric_names()

    # apply the lookback window
    if self.reduction_window is not None:
        self.data = self.data.tail(self.reduction_window)

    self.param_columns = [col for col in self.data.columns if col not in metric_names()]
    self.param_table = self.data[self.param_columns]
    self.param_table.insert(0, self.reduction_metric, self.data[self.reduction_metric])

    return self



from .scan_prepare import scan_prepare
from .scan_run import scan_run


class Scan:
    """Suite of operations for training and evaluating Keras neural networks.

    Inputs train/dev data and a set of parameters as a dictionary. The name and
    experiment number must also be chosen since they define the output
    filenames. The model must also be specified of the form

        my_model(x_train, y_train, x_val, y_val, params),

    and the dictionary

        d = {
            'fcc_layer_1_N': [50, 100, 200],
            'fcc_layer_1_act': ['relu', 'tanh'],
            'fcc_layer_1_dropout': (0, 0.1, 5)    # 5 points between 0 and 0.1
        }

    The dictionary is parsed for every run and only one entry per parameter
    is fed into the neural network at a time.

    Important note: the user has two options when specifying input data.

    Option 1:
        Specify x, y and val_split. The training and validation data mixture
        (x, y) will be randomly split into the training and validation datasets
        as per the split specified in val_split.

    Option 2:
        Specify x, y and x_val, y_val. This would allow the user to specify
        their own validation datasets. Keras by default shuffles data during
        training, so the user need only be sure that the split specified is
        correct. This allows for not only reproducibility, but randomizing the
        data on the user's own terms. This is critical if the user wishes to
        augment their training data without augmenting their validation data
        (which is the only acceptable practice!).


    Parameters
    ----------
    x : ndarray
        1d or 2d array consisting of the training data. `x` should have the
        shape (m, n), where m is the number of training examples and n is the
        number of features. Extra dimensions can be added to account for the
        channels entry in convolutional neural networks.
    y : ndarray
        The labels corresponding to the training data. `y` should have the
        shape (m, c) where c is the number of classes. A binary classification
        problem will have c=1.
    params : python dictionary
        Lists all permutations of hyperparameters, a subset of which will be
        selected at random for training and evaluation.
    dataset_name : str
        References the name of the experiment. The dataset_name and
        experiment_no will be concatenated to produce the file name for the
        results saved in the local directory.
    experiment_no : str
        Indexes the user's choice of experiment number.
    model : keras_model
        A Keras style model which compiles and fits the data, and returns
        the history and compiled model.
    val_split : float, optional
        The proportion of the input `x` which is set aside as the
        cross-validation data. (Default is 0.3).
    shuffle : bool, optional
        If True, shuffle the data in x and y before splitting into the train
        and cross-validation datasets. (Default is True).
    random_method : uniform, stratified, lhs, lhs_sudoku
        Determinines the way in which the grid_downsample is applied. The
        default setting is 'uniform'.
    search_method : {None, 'random', 'linear', 'reverse'}
        Determines the random sampling of the dictionary. `random` picks one
        hyperparameter point at random and removes it from the list, then
        samples again. `linear` starts from the start of the grid and moves
        forward, and `reverse` starts at the end of the grid and moves
        backwards.
    reduction_method : {None, 'spear'}
        Method for honing in on the optimal hyperparameter subspace. (Default
        is None).
    reduction_interval : int
        The number of reduction method rounds that will be performed. (Default
        is None).
    reduction_window : int
        The number of rounds of the reduction method before observing the
        results. (Default is None).
    grid_downsample : int
        The fraction of `params` that will be tested (Default is None).
    reduction_metric : {'val_acc'}
        Metric used to tune the reductions.
    talos_log_name : str
        The lame of the saved Talos log. (Default is 'talos.log').
    debug : bool
        Implements debugging feedback. (Default is False).
    x_val : ndarray
        User specified cross-validation data. (Default is None).
    y_val : ndarray
        User specified cross-validation labels. (Default is None).
    print_params : bool
        Print params for each round on screen (useful when using TrainingLog
        callback for visualization)

    """

    # TODO: refactor this so that we don't initialize global variables
    global self

    def __init__(self, x, y, params, model,
                 dataset_name=None, experiment_no=None,
                 x_val=None, y_val=None,
                 val_split=.3, shuffle=True,
                 round_limit=None,
                 grid_downsample=None,
                 random_method='uniform_mersenne',
                 seed=None,
                 search_method='random',
                 reduction_method=None,
                 reduction_interval=50,
                 reduction_window=20,
                 reduction_threshold=0.2,
                 reduction_metric='val_acc',
                 reduce_loss=False,
                 last_epoch_value=False,
                 talos_log_name='talos.log',
                 clear_tf_session=True,
                 functional_model=False,
                 disable_progress_bar=False,
                 print_params=False,
                 debug=False,):

        # NOTE: these need to be follow the order from __init__
        # and all paramaters needs to be included here and only here.

        self.x = x
        self.y = y
        self.params = params
        self.dataset_name = dataset_name
        self.experiment_no = experiment_no
        self.model = model
        self.x_val = x_val
        self.y_val = y_val
        self.val_split = val_split
        self.shuffle = shuffle
        self.random_method = random_method
        self.search_method = search_method
        self.reduction_method = reduction_method
        self.reduction_interval = reduction_interval
        self.reduction_window = reduction_window
        self.grid_downsample = grid_downsample
        self.reduction_threshold = reduction_threshold
        self.reduction_metric = reduction_metric
        self.reduce_loss = reduce_loss
        self.round_limit = round_limit
        self.talos_log_name = talos_log_name
        self.debug = debug
        self.seed = seed
        self.clear_tf_session = clear_tf_session
        self.disable_progress_bar = disable_progress_bar
        self.functional_model = functional_model
        self.last_epoch_value = last_epoch_value
        self.print_params = print_params
        # input parameters section ends

        self._null = self.runtime()

    def runtime(self):

        self = scan_prepare(self)
        self = scan_run(self)



def reduce_finish(self):

    '''Takes input from a Reducer in form of a tuple
    where the values the hyperparamater name and the
    value to drop. Returns self with a modified param_log.'''

    # get the column index

    to_remove_col = self.param_reference[self._reduce_keys[1]]

    value_to_remove = self._reduce_keys[0]

    # pick the index numbers for dropping available permutations
    indexs_to_drop = self.param_grid[self.param_grid[:, to_remove_col] == value_to_remove][:,-1]

    # drop the index numbers
    self.param_log = list(set(self.param_log).difference(set(indexs_to_drop)))

    return self



def classify(y):

    '''Detects if prediction is binary, multi-label or multi-class'''

    shape = detect_shape(y)

    if shape > 1:
        return 'multi_class'

    elif y.max() <= 1:
        return 'binary_class'
    else:
        return 'multi_label'


def detect_shape(y):

    try:
        return y.shape[1]
    except IndexError:
        return 1



from numpy import arange, unique, array, column_stack
from itertools import product

from ..reducers.sample_reducer import sample_reducer


class ParamGrid:

    '''Suite for handling parameters internally within Talos

    Takes as input the parameter dictionary from the user, and
    returns a class object which can then be used to pick parameters
    for each round together with other parameter related operations.

    '''

    def __init__(self, main_self):

        self.main_self = main_self

        # convert the input to useful format
        self._p = self._param_input_conversion()

        # build the parameter permutation grid
        self.param_grid = self._param_grid()

        # reduce according to downsample
        if self.main_self.grid_downsample is not None:
            self.param_grid = sample_reducer(self)

        # create a index for logging purpose
        self.param_log = list(range(len(self.param_grid)))

        # add the log index to param grid
        self.param_grid = column_stack((self.param_grid, self.param_log))

    def _param_grid(self):

        '''CREATE THE PARAMETER PERMUTATIONS

        This is done once before starting the experiment.
        Takes in the parameter dictionary, and returns
        every possible permutation in an array.
        '''

        ls = [list(self._p[key]) for key in self._p.keys()]
        _param_grid_out = array(list(product(*ls)))

        return _param_grid_out

    def _param_input_conversion(self):

        '''DETECT PARAM FORMAT

        Checks of the hyperparameter input format is list
        or tupple in the params dictionary and expands accordingly.

        '''

        out = {}

        for param in self.main_self.params.keys():

            # for range/step style input
            if isinstance(self.main_self.params[param], tuple):
                out[param] = self._param_range(self.main_self.params[param][0],
                                               self.main_self.params[param][1],
                                               self.main_self.params[param][2])
            # all other input styles
            else:
                out[param] = self.main_self.params[param]

        return out

    def _param_range(self, start, end, n):

        '''PARAMETER RANGE

        Deals with the format where a start, end
        and steps values are given for a parameter
        in a tuple format.

        This is called internally from param_format()
        '''

        try:
            out = arange(start, end, (end - start) / n, dtype=float)
        # this is for python2
        except ZeroDivisionError:
            out = arange(start, end, (end - start) / float(n), dtype=float)

        if type(start) == int and type(end) == int:
            out = out.astype(int)
            out = unique(out)

        return out



from numpy import mean, std

from sklearn.metrics import f1_score

from ..utils.validation_split import kfold
from ..utils.best_model import best_model, activate_model


class Evaluate:

    '''Class for evaluating models based on the Scan() object'''

    def __init__(self, scan_object):

        '''Takes in as input a Scan() object'''

        self.scan_object = scan_object
        self.data = scan_object.data

    def evaluate(self, x, y,
                 model_id=None,
                 folds=5,
                 shuffle=True,
                 average='binary',
                 metric='val_acc',
                 asc=False):

        '''Evaluate model against f1-score'''

        out = []
        if model_id is None:
            model_id = best_model(self.scan_object, metric, asc)

        model = activate_model(self.scan_object, model_id)

        kx, ky = kfold(x, y, folds, shuffle)

        for i in range(folds):
            y_pred = model.predict(kx[i]) >= 0.5
            scores = f1_score(y_pred, ky[i], average=average)
            out.append(scores * 100)

        print("%.2f%% (+/- %.2f%%)" % (mean(out), std(out)))



import pandas as pd
from numpy import nan
from keras.utils import to_categorical


base = 'https://raw.githubusercontent.com/autonomio/datasets/master/autonomio-datasets/'


def limit_rows(data, samples):

    return data.sample(frac=1).head(samples)


def icu_mortality(samples=None):

    df = pd.read_csv(base + 'icu_mortality.csv')
    df = df.dropna(thresh=3580, axis=1)
    df = df.dropna()
    df = limit_rows(df, samples)
    y = df['hospitalmortality'].astype(int).values
    x = df.drop('hospitalmortality', axis=1).values

    return x, y


def titanic():

    df = pd.read_csv(base + 'titanic.csv')

    y = df.survived.values

    x = df[['age', 'sibsp', 'parch']]
    cols = ['class', 'embark_town', 'who', 'deck', 'sex']

    for col in cols:
        x = pd.merge(x,
                     pd.get_dummies(df[col]),
                     left_index=True,
                     right_index=True)

    x = x.dropna()
    x = x.values

    return x, y


def iris():

    df = pd.read_csv(base + 'iris.csv')
    df['species'] = df['species'].factorize()[0]
    df = df.sample(len(df))
    y = to_categorical(df['species'])
    x = df.iloc[:, :-1].values

    y = to_categorical(df['species'])
    x = df.iloc[:, :-1].values

    return x, y


def cervical_cancer():

    df = pd.read_csv(base + 'cervical_cancer.csv')
    df = df.replace('?', nan)
    df = df.drop(['citology', 'hinselmann', 'biopsy'], axis=1)
    df = df.drop(['since_first_diagnosis',
                  'since_last_diagnosis'], axis=1).dropna()

    df = df.astype(float)

    y = df.schiller.values
    x = df.drop('schiller', axis=1).values

    return x, y


def breast_cancer():

    df = pd.read_csv(base + 'breast_cancer.csv')

    # then some minimal data cleanup
    df.drop("Unnamed: 32", axis=1, inplace=True)
    df.drop("id", axis=1, inplace=True)

    # separate to x and y
    y = df.diagnosis.values
    x = df.drop('diagnosis', axis=1).values

    # convert the string labels to binary
    y = (y == 'M').astype(int)

    return x, y



import pandas as pd


def correlation(self):

    '''Correlation Reducers

    Note that this set of reducers works only for the continuous
    and stepped (e.g. batch size) hyperparameters.

    '''

    out = self.param_table.corr(method='spearman')[self.reduction_metric]
    out = out.dropna()

    if len(out) == 0:
        self._reduce_keys = None
        return self

    out = out[1:].sort_values(ascending=False)
    out = out.index[-1], out[-1]

    if abs(out[1]) >= self.reduction_threshold:
        dummy_cols = pd.get_dummies(self.param_table[out[0]])
        dummy_cols.insert(0,
                          self.reduction_metric,
                          self.param_table[self.reduction_metric])

    # case where threshold is not met
    else:
        self._reduce_keys = None
        return self

    # all other cases continue
    to_drop_temp = dummy_cols.corr(method='spearman')[self.reduction_metric]

    # pick the drop method based on paramaters
    if self.reduce_loss is False:
        self._reduce_keys = to_drop_temp.sort_values().index[0], out[0]
    else:
        self._reduce_keys = to_drop_temp.sort_values().index[-2], out[0]

    return self



import os
import pandas as pd
import shutil
import numpy as np

from ..utils.best_model import best_model, activate_model

# NOTE: this has some overlap with code in evaluate.py
# and needs to be cleaned up.
# TODO: needs to also deploy the hyperparameter configuration
# and some kind of summary of the experiment and then finally
# pack everything into a zip file


class Deploy:

    '''Functionality for deploying a model to a filename'''

    def __init__(self, scan_object, model_name, metric='val_acc', asc=False):

        self.scan_object = scan_object
        os.mkdir(model_name)
        self.path = model_name + '/' + model_name
        self.model_name = model_name
        self.metric = metric
        self.asc = asc
        self.data = scan_object.data
        self.best_model = best_model(scan_object, metric, asc)
        self.model = activate_model(scan_object, self.best_model)

        # runtime
        self.save_model_as()
        self.save_details()
        self.save_data()
        self.save_results()
        self.save_params()
        self.save_readme()
        self.package()

    def save_model_as(self):

        '''Model Saver
        WHAT: Saves a trained model so it can be loaded later
        for predictions by predictor().
        '''

        model_json = self.model.to_json()
        with open(self.path + "_model.json", "w") as json_file:
            json_file.write(model_json)

        self.model.save_weights(self.path + "_model.h5")
        print("Deploy package" + " " + self.model_name + " " + "have been saved.")

    def save_details(self):

        self.scan_object.details.to_csv(self.path + '_details.txt')

    def save_data(self):

        x = pd.DataFrame(self.scan_object.x[:100])
        y = pd.DataFrame(self.scan_object.y[:100])

        x.to_csv(self.path + '_x.csv', header=None, index=None)
        y.to_csv(self.path + '_y.csv', header=None, index=None)

    def save_results(self):

        self.scan_object.data.to_csv(self.path + '_results.csv')

    def save_params(self):

        np.save(self.path + '_params', self.scan_object.params)

    def save_readme(self):

        txt = 'To activate the assets in the Talos deploy package: \n\n   from talos.commands.restore import Restore \n   a = Restore(\'path_to_asset\')\n\nNow you will have an object similar to the Scan object, which can be used with other Talos commands as you would be able to with the Scan object'

        text_file = open(self.path.split('/')[0] + '/README.txt', "w")
        text_file.write(txt)
        text_file.close()

    def package(self):

        shutil.make_archive(self.model_name, 'zip', self.model_name)
        shutil.rmtree(self.model_name)



def metric_names():

    '''These are used as a shorthand for filtering out columns
    that should not be included as depedent variables for optimizing.'''

    return ['round_epochs',
            'loss',
            'val_loss',
            'acc',
            'val_acc',
            'fmeasure_acc',
            'val_fmeasure_acc',
            'recall_acc',
            'val_recall_acc',
            'precision_acc',
            'val_precision_acc',
            'matthews_correlation_acc',
            'val_matthews_correlation_acc'
            ]



#! /usr/bin/env python
#
# Copyright (C) 2018 Mikko Kotila

DESCRIPTION = "Talos Hyperparameter Scanner for Keras"
LONG_DESCRIPTION = """\
Talos provides a hyperparameter scanning solution that
allows using any Keras model as they are, with the simple
change that instead of calling the parameter (e.g. epochs=25),
you call it from a dictionary with an identical label (e.g. params['epochs']).
"""

DISTNAME = 'talos'
MAINTAINER = 'Mikko Kotila'
MAINTAINER_EMAIL = 'mailme@mikkokotila.com'
URL = 'http://autonom.io'
LICENSE = 'MIT'
DOWNLOAD_URL = 'https://github.com/autonomio/talos/'
VERSION = '0.4.3'

try:
    from setuptools import setup
    _has_setuptools = True
except ImportError:
    from distutils.core import setup


def check_dependencies():

    install_requires = []

    try:
        import numpy
    except ImportError:
        install_requires.append('numpy')
    try:
        import pandas
    except ImportError:
        install_requires.append('pandas')
    try:
        import tensorflow
    except ImportError:
        install_requires.append('tensorflow')
    try:
        import keras
    except ImportError:
        install_requires.append('keras')
    try:
        import astetik
    except ImportError:
        install_requires.append('astetik')
    try:
        import sklearn
    except ImportError:
        install_requires.append('sklearn')
    try:
        import tqdm
    except ImportError:
        install_requires.append('tqdm')
    try:
        import chances
    except ImportError:
        install_requires.append('chances')
    try:
        import kerasplotlib
    except ImportError:
        install_requires.append('kerasplotlib')
    try:
        import wrangle
    except ImportError:
        install_requires.append('wrangle')

    return install_requires


if __name__ == "__main__":

    install_requires = check_dependencies()

    setup(name=DISTNAME,
          author=MAINTAINER,
          author_email=MAINTAINER_EMAIL,
          maintainer=MAINTAINER,
          maintainer_email=MAINTAINER_EMAIL,
          description=DESCRIPTION,
          long_description=LONG_DESCRIPTION,
          license=LICENSE,
          url=URL,
          version=VERSION,
          download_url=DOWNLOAD_URL,
          install_requires=install_requires,
          packages=['talos',
                    'talos.scan',
                    'talos.examples',
                    'talos.utils',
                    'talos.model',
                    'talos.parameters',
                    'talos.reducers',
                    'talos.metrics',
                    'talos.commands'],

          classifiers=[
                 'Intended Audience :: Science/Research',
                 'Programming Language :: Python :: 3.6',
                 'License :: OSI Approved :: MIT License',
                 'Topic :: Scientific/Engineering :: Human Machine Interfaces',
                 'Topic :: Scientific/Engineering :: Artificial Intelligence',
                 'Topic :: Scientific/Engineering :: Mathematics',
                 'Operating System :: POSIX',
                 'Operating System :: Unix',
                 'Operating System :: MacOS'],)



from numpy import random


def round_params(self):

    '''Picks the paramaters for a round based on the available
    paramater permutations using the param_log index'''

    # pick the permutation for the round
    if self.search_method == 'random':
        _choice = random.choice(self.param_log)

    elif self.search_method == 'linear':
        _choice = min(self.param_log)

    elif self.search_method == 'reverse':
        _choice = max(self.param_log)

    # remove the current choice from permutations
    self.param_log.remove(_choice)

    # create a dictionary for the current round
    _round_params_dict = {}
    x = 0
    for key in self.param_reference.keys():
        _round_params_dict[key] = self.param_grid[_choice][x]
        x += 1

    return _round_params_dict



from tqdm import tqdm

from ..utils.results import result_todf, peak_epochs_todf
from .scan_round import scan_round
from .scan_finish import scan_finish


def scan_run(self):

    '''The high-level management of the scan procedures
    onwards from preparation. Manages round_run()'''

    # enforce round_limit
    self.param_grid = self.param_grid[:self.round_limit]

    # main loop for the experiment
    # NOTE: the progress bar is also updated on line 73
    self.pbar = tqdm(total=len(self.param_log),
                     disable=self.disable_progress_bar)
    while len(self.param_log) != 0:
        self = scan_round(self)
        self.pbar.update(1)
    self.pbar.close()

    # save the results
    self = result_todf(self)
    self.peak_epochs_df = peak_epochs_todf(self)
    self._null = self.logfile.close()

    self = scan_finish(self)

    print('Scan Finished!')



from .reduce_prepare import reduce_prepare
from .reduce_finish import reduce_finish
from .correlation import correlation


def reduce_run(self):

    '''The process run script for reduce
    procedures; takes care of everything
    related with reduction. When new
    reduction methods are added, they need
    to be added as options here.
    '''

    # prepare log for reduction analysis
    self = reduce_prepare(self)

    # run the selected reduction method
    if self.reduction_method == 'correlation':
        self = correlation(self)

    # TODO: the case where reduction_method
    # is not selected or is wrong could be
    # handled better.

    # handle the dropping of permutations
    if self._reduce_keys is None:
        return self
    else:
        return reduce_finish(self)



from pandas import read_csv
from astetik import line, hist, corr, regs, bargrid, kde
from ..metrics.names import metric_names


class Reporting:

    '''A suite of commands that are useful for analyzing the results
    of a completed scan, or during a scan.

    filename :: the name of the experiment log from Scan()'''

    def __init__(self, source=None):

        '''Takes as input a filename to the experiment
        log or the Scan object'''

        if isinstance(source, str):
            self.data = read_csv(source)
        else:
            self.data = source.data

    def high(self, metric='val_acc'):

        '''Returns the highest value for a given metric'''

        return max(self.data[metric])

    def rounds(self):

        '''Returns the number of rounds in the experiment'''

        return len(self.data)

    def rounds2high(self, metric='val_acc'):

        '''Returns the number of rounds it took to get to the
        highest value for a given metric.'''

        return self.data[self.data[metric] == self.data[metric].max()].index[0]

    def low(self, metric='val_acc'):

        '''Returns the minimum value for a given metric'''

        return min(self.data[metric])

    def correlate(self, metric='val_acc'):

        '''Returns a correlation table against a given metric. Drops
        all other metrics and correlates against hyperparameters only.'''

        columns = [c for c in self.data.columns if c not in metric_names()]
        out = self.data[columns]
        out.insert(0, metric, self.data[metric])
        out = out.corr()[metric]

        return out[out != 1]

    def plot_line(self, metric='val_acc'):

        '''A line plot for a given metric where rounds is on x-axis

        NOTE: remember to invoke %matplotlib inline if in notebook

        metric :: the metric to correlate against

        '''

        return line(self.data, metric)

    def plot_hist(self, metric='val_acc', bins=10):

        '''A histogram for a given metric

        NOTE: remember to invoke %matplotlib inline if in notebook

        metric :: the metric to correlate against
        bins :: number of bins to use in histogram

        '''

        return hist(self.data, metric, bins=bins)

    def plot_corr(self, metric='val_acc', color_grades=5):

        '''A heatmap with a single metric and hyperparameters.

        NOTE: remember to invoke %matplotlib inline if in notebook

        metric :: the metric to correlate against
        color_grades :: number of colors to use in heatmap'''

        cols = self._cols(metric)

        return corr(self.data[cols], color_grades=color_grades)

    def plot_regs(self, x='val_acc', y='val_loss'):

        '''A regression plot with data on two axis

        x = data for the x axis
        y = data for the y axis
        '''

        return regs(self.data, x, y)

    def plot_bars(self, x, y, hue, col):

        '''A comparison plot with 4 axis'''

        return bargrid(self.data,
                       x=x,
                       y=y,
                       hue=hue,
                       col=col,
                       col_wrap=4)

    def plot_kde(self, x, y=None):

        '''Kernel Destiny Estimation type histogram with
        support for 1 or 2 axis of data'''

        return kde(self.data, x, y)

    def table(self, metric='val_acc', sort_by=None, ascending=False):

        '''Shows a table with hyperparameters and a given metric

        EXAMPLE USE:

        ra1 = Reporting('diabetes_1.csv')
        ra1.table(sort_by='fmeasure_acc', ascending=False)

        PARAMS:

        metric :: accepts single column name as string or multiple in list
        sort_by :: the colunm name sorting should be based on
        ascending :: if sorting is ascending or not

        '''

        cols = self._cols(metric)

        if sort_by is None:
            sort_by = metric

        out = self.data[cols].sort_values(sort_by, ascending=ascending)

        return out

    def best_params(self, metric='val_acc', n=10):

        '''Get the best parameters of the experiment based on a metric.
        Returns a numpy array with the values in a format that can be used
        with the talos backend in Scan(). Adds an index as the last column.'''

        cols = self._cols(metric)
        out = self.data[cols].sort_values(metric, ascending=False)
        out = out.drop(metric, axis=1).head(n)
        out.insert(out.shape[1], 'index_num', range(len(out)))

        return out.values

    def _cols(self, metric):

        '''Helper to remove other than desired metric from data table'''

        cols = [col for col in self.data.columns if col not in metric_names()]

        if isinstance(metric, list) is False:
            metric = [metric]
        for i, metric in enumerate(metric):
            cols.insert(i, metric)

        return cols
